@@ -3,13 +3,13 @@ from __future__ import annotations
 from dataclasses import asdict, is_dataclass
 from enum import auto, Enum
 from functools import wraps
-from typing import Any, Callable, cast, Type, TYPE_CHECKING, Union
+from typing import Any, Callable, cast, Optional, Type, TYPE_CHECKING, Union
 
 from pydantic import BaseModel, ValidationError
 from pydantic.schema import model_schema
-from quart import request
-from quart.exceptions import BadRequest
+from quart import current_app, request
 from werkzeug.datastructures import Headers
+from werkzeug.exceptions import BadRequest
 
 if TYPE_CHECKING:
     from pydantic.dataclasses import Dataclass
@@ -18,7 +18,6 @@ if TYPE_CHECKING:
 QUART_SCHEMA_REQUEST_ATTRIBUTE = "_quart_schema_request_schema"
 QUART_SCHEMA_RESPONSE_ATTRIBUTE = "_quart_schema_response_schemas"
 QUART_SCHEMA_QUERYSTRING_ATTRIBUTE = "_quart_schema_querystring_schema"
-REF_PREFIX = "#/components/schemas/"
 
 
 class SchemaInvalidError(Exception):
@@ -26,11 +25,14 @@ class SchemaInvalidError(Exception):
 
 
 class ResponseSchemaValidationError(Exception):
-    pass
+    def __init__(self, validation_error: Optional[ValidationError] = None) -> None:
+        self.validation_error = validation_error
 
 
 class RequestSchemaValidationError(BadRequest):
-    pass
+    def __init__(self, validation_error: Union[TypeError, ValidationError]) -> None:
+        super().__init__()
+        self.validation_error = validation_error
 
 
 class DataSource(Enum):
@@ -50,23 +52,22 @@ def validate_querystring(model_class: Union[Type[BaseModel], Type[Dataclass]]) -
             a class that inherits from pydantic's BaseModel. All the
             fields must be optional.
     """
-    schema = model_schema(model_class, ref_prefix=REF_PREFIX)
+    schema = model_schema(model_class)
 
     if len(schema.get("required", [])) != 0:
         raise SchemaInvalidError("Fields must be optional")
 
     def decorator(func: Callable) -> Callable:
-        setattr(func, QUART_SCHEMA_QUERYSTRING_ATTRIBUTE, schema)
+        setattr(func, QUART_SCHEMA_QUERYSTRING_ATTRIBUTE, model_class)
 
         @wraps(func)
         async def wrapper(*args: Any, **kwargs: Any) -> Any:
             try:
                 model = model_class(**request.args)
-            except (TypeError, ValidationError):
-                raise RequestSchemaValidationError()
+            except (TypeError, ValidationError) as error:
+                raise RequestSchemaValidationError(error)
             else:
-                return await func(*args, query_args=model, **kwargs)
-            return await func(*args, **kwargs)
+                return await current_app.ensure_async(func)(*args, query_args=model, **kwargs)
 
         return wrapper
 
@@ -91,14 +92,14 @@ def validate_request(
         source: The source of the data to validate (json or form
             encoded).
     """
-    schema = model_schema(model_class, ref_prefix=REF_PREFIX)
+    schema = model_schema(model_class)
     if source == DataSource.FORM and any(
         schema["properties"][field]["type"] == "object" for field in schema["properties"]
     ):
-        raise SchemaInvalidError("Form must have nested objects")
+        raise SchemaInvalidError("Form must not have nested objects")
 
     def decorator(func: Callable) -> Callable:
-        setattr(func, QUART_SCHEMA_REQUEST_ATTRIBUTE, (schema, source))
+        setattr(func, QUART_SCHEMA_REQUEST_ATTRIBUTE, (model_class, source))
 
         @wraps(func)
         async def wrapper(*args: Any, **kwargs: Any) -> Any:
@@ -109,10 +110,10 @@ def validate_request(
 
             try:
                 model = model_class(**data)
-            except (TypeError, ValidationError):
-                raise RequestSchemaValidationError()
+            except (TypeError, ValidationError) as error:
+                raise RequestSchemaValidationError(error)
             else:
-                return await func(*args, data=model, **kwargs)
+                return await current_app.ensure_async(func)(*args, data=model, **kwargs)
 
         return wrapper
 
@@ -137,16 +138,15 @@ def validate_response(
         status_code: The status code this validation applies
             to. Defaults to 200.
     """
-    schema = model_schema(model_class, ref_prefix=REF_PREFIX)
 
     def decorator(func: Callable) -> Callable:
         schemas = getattr(func, QUART_SCHEMA_RESPONSE_ATTRIBUTE, {})
-        schemas[status_code] = schema
+        schemas[status_code] = model_class
         setattr(func, QUART_SCHEMA_RESPONSE_ATTRIBUTE, schemas)
 
         @wraps(func)
         async def wrapper(*args: Any, **kwargs: Any) -> Any:
-            result = await func(*args, **kwargs)
+            result = await current_app.ensure_async(func)(*args, **kwargs)
 
             status_or_headers = None
             headers = None
@@ -161,12 +161,12 @@ def validate_response(
             ):
                 status = int(status_or_headers)
 
-            if status in schemas:
+            if status == status_code:
                 if isinstance(value, dict):
                     try:
                         model_value = model_class(**value)
-                    except ValidationError:
-                        raise ResponseSchemaValidationError()
+                    except ValidationError as error:
+                        raise ResponseSchemaValidationError(error)
                 elif type(value) == model_class:
                     model_value = value
                 else:
